@@ -61,7 +61,8 @@
 extern int _sigaction_nobind (int sig, const struct sigaction *nsv, struct sigaction *osv);
 #endif
 
-static int sigvec__(signo, sv, osv, bind)
+static int
+sigvec__(signo, sv, osv, bind)
 	int signo;
 	struct sigvec *sv, *osv;
 	int bind;
@@ -84,7 +85,8 @@ static int sigvec__(signo, sv, osv, bind)
 	return (ret);
 }
 
-int sigvec(signo, sv, osv)
+int
+sigvec(signo, sv, osv)
         int signo;
         struct sigvec *sv, *osv;
 {
@@ -92,7 +94,8 @@ int sigvec(signo, sv, osv)
 }
 
 #if defined(__DYNAMIC__)
-int _sigvec_nobind(signo, sv, osv)
+int
+_sigvec_nobind(signo, sv, osv)
         int signo;
         struct sigvec *sv, *osv;
 {
@@ -100,7 +103,8 @@ int _sigvec_nobind(signo, sv, osv)
 }
 #endif
 
-int sigsetmask(mask)
+int
+sigsetmask(mask)
 	int mask;
 {
 	int omask, n;
@@ -111,7 +115,8 @@ int sigsetmask(mask)
 	return (omask);
 }
 
-int sigblock(mask)
+int
+sigblock(mask)
 	int mask;
 {
 	int omask, n;
@@ -122,32 +127,79 @@ int sigblock(mask)
 	return (omask);
 }
 
-int sigpause(mask)
+int
+sigpause(mask)
 	int mask;
 {
 	return (sigsuspend((sigset_t *)&mask));
 }
 
-int sighold(sig)
+int
+sighold(sig)
 	int sig;
 {
 	sigset_t mask;
 
-	if ((sig < 0) || (sig > NSIG))
-		return(EINVAL);
+	if ((sig <= 0) || (sig >= NSIG)) {
+		errno = EINVAL;
+		return(-1);
+	}
 	sigemptyset(&mask);
 	sigaddset(&mask, sig);
 	return(sigprocmask(SIG_BLOCK, &mask,(sigset_t *)0));
 }
-int sigrelse(sig)
+
+int
+sigrelse(sig)
 	int sig;
 {
 	sigset_t mask;
 
-	if ((sig < 0) || (sig > NSIG))
-		return(EINVAL);
+	if ((sig <= 0) || (sig >= NSIG)) {
+		errno = EINVAL;
+		return(-1);
+	}
 	sigemptyset(&mask);
 	sigaddset(&mask, sig);
 	return(sigprocmask(SIG_UNBLOCK, &mask,(sigset_t *)0));
 }
 
+
+int
+sigignore(sig)
+	int sig;
+{
+	return (signal(sig, SIG_IGN) == SIG_ERR ? -1 : 0);
+}
+
+void (*sigset(int sig, void (*disp)(int)))(int) {
+	sigset_t omask;
+	int blocked;
+	struct sigaction oact;
+
+	if ((sig <= 0) || (sig >= NSIG)) {
+		errno = EINVAL;
+		return (SIG_ERR);
+	}
+	if (-1 == sigprocmask(0, NULL, &omask))
+		return (SIG_ERR);
+	blocked = sigismember(&omask, sig);
+	if (disp == SIG_HOLD) {
+		if (blocked)
+			return (SIG_HOLD);
+		if ((-1 == sigaction(sig, NULL, &oact)) ||
+		    (-1 == sighold(sig)))
+			return (SIG_ERR);
+		return (sig_t)oact.sa_handler;
+	} else {
+		if (blocked) {
+			if (-1 == sigrelse(sig))
+				return (SIG_ERR);
+		}
+		sig_t rv = signal(sig, disp);
+		if (rv != SIG_ERR)
+			return blocked ? SIG_HOLD : rv;
+		else
+			return (rv);
+	}
+}
